/* ============================================================================
 * AMES Wholesale Power Market Test Bed (Java): A Free Open-Source Test-Bed
 *         for the Agent-based Modeling of Electricity Systems
 * ============================================================================
 *
 * (C) Copyright 2013, by Hongyan Li, Junjie Sun, Sean L. Mooney and Leigh Tesfatsion
 *
 *    Homepage: http://www.econ.iastate.edu/tesfatsi/testMarketHome.htm
 *
 * LICENSING TERMS
 * The AMES Market Package is licensed by the copyright holders (Junjie Sun,
 * Hongyan Li, Sean L. Mooney, and Leigh Tesfatsion) as free open-source software under the
 * terms of the GNU General Public License (GPL). Anyone who is interested is
 * allowed to view, modify, and/or improve upon the code used to produce this
 * package, but any software generated using all or part of this code must be
 * released as free open-source software in turn. The GNU GPL can be viewed in
 * its entirety as in the following site: http://www.gnu.org/licenses/gpl.html
 */
package amesmarket;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Scanner;

import amesmarket.CaseFileData.GenData;


/**
 * Parse a case file.
 * @author Sean L. Mooney
 *
 */
public class CaseFileReader {

    //TOKENS
    private static final String BASE_S = "BASE_S";
    private static final String BASE_V = "BASE_V";
    private static final String NODE_DATA_START = "#NodeDataStart";
    private static final String NODE_DATA_END = "#NodeDataEnd";
    private static final String BRANCH_DATA_START = "#BranchDataStart";
    private static final String BRANCH_DATA_END = "#BranchDataEnd";
    private static final String GEN_DATA_START = "#GenDataStart";
    private static final String GEN_DATA_END = "#GenDataEnd";
    private static final String LSE_DATA_FIXED_DEM_START = "#LSEDataFixedDemandStart";
    private static final String LSE_DATA_FIXED_DEM_END = "#LSEDataFixedDemandEnd";
    private static final String LSE_DATA_PRICE_SENS_DEM_START = "#LSEDataPriceSensitiveDemandStart";
    private static final String LSE_DATA_PRICE_SENS_DEM_END = "#LSEDataPriceSensitiveDemandEnd";
    private static final String LSE_DATA_HYBRID_DEM_START = "#LSEDataHybridDemandStart";
    private static final String LSE_DATA_HYBRID_DEM_END = "#LSEDataHybridDemandEnd";
    private static final String GEN_LEARNING_DATA_START = "#GenLearningDataStart";
    private static final String GEN_LEARNING_DATA_END = "#GenLearningDataEnd";


    private static final String WS_REG_EX = "\\s+";

    private SimpleLineReader inputReader;
    private String currentLine = null;

    /**
     * Read in/load the test file configuration.
     *
     * @param testCaseFile
     * @throws BadDataFileFormatException
     */
    public CaseFileData loadCaseFileData(final File testCaseFile) throws FileNotFoundException, IOException, BadDataFileFormatException {
      inputReader = new SimpleLineReader(testCaseFile);
      return loadCaseFileData();
    }

    public CaseFileData loadCaseFileData(final Reader testCaseInput) throws BadDataFileFormatException{
      inputReader = new SimpleLineReader(testCaseInput);
      return loadCaseFileData();
    }

    private boolean move(){
        currentLine = inputReader.nextLine();
        return currentLine != null;
    }

    private void match(String expected) throws BadDataFileFormatException{
        if(!expected.equals(currentLine)) throw new BadDataFileFormatException(
                "Expected " + expected + ". Found" + currentLine);
    }

    private void parseDataFile(CaseFileData testConf) throws BadDataFileFormatException{

        boolean hasGenLearningData = false;

        while (inputReader.hasNext() && move()) {
            if(currentLine.startsWith(BASE_S)){
                parseBASE_S(testConf);
            }else if(currentLine.startsWith(BASE_V)){
                parseBASE_V(testConf);
            }else if(currentLine.equals(NODE_DATA_START)){
                parseNodeData(testConf);
            }else if(currentLine.equals(BRANCH_DATA_START)){
                parseBranchData(testConf);
            }else if(currentLine.equals(GEN_DATA_START)){
                parseGenData(testConf);
            }else if(currentLine.equals(LSE_DATA_FIXED_DEM_START)){
                parseLSEFixedDemand(testConf);
            }else if(currentLine.equals(LSE_DATA_PRICE_SENS_DEM_START)){
                parseLSEPSensDemand(testConf);
            }else if(currentLine.equals(LSE_DATA_HYBRID_DEM_START)){
                parseLSEHybDemand(testConf);
            }else if(currentLine.equals(GEN_LEARNING_DATA_START)){
                hasGenLearningData = true;
                parseGenLearningData(testConf);
            }
        }

        if(!hasGenLearningData){
            loadDefaultGenLearningData(testConf);
        }
    }

    private void parseBASE_S(CaseFileData testConf) throws BadDataFileFormatException{
        //Split the string on white space
        String[] splits = currentLine.split("\\s+");
        if(splits.length != 2){
            throw new BadDataFileFormatException(currentLine);
        }
        testConf.baseS = Double.parseDouble(splits[1]);
    }

    private void parseBASE_V(CaseFileData testConf) throws BadDataFileFormatException{
        //Split the string on white space
        String[] splits = currentLine.split(WS_REG_EX);
        if(splits.length != 2){
            throw new BadDataFileFormatException(currentLine);
        }
        testConf.baseV = Double.parseDouble(splits[1]);
    }

    private void parseNodeData(CaseFileData testConf) throws BadDataFileFormatException{
        testConf.nodeData = new Object[1][2];
        move();

        String[] splits = currentLine.split(WS_REG_EX);

        if(splits.length != 2){
            throw new BadDataFileFormatException(currentLine);
        }

        int  numNodes = Integer.parseInt(splits[0]);
        testConf.iNodeData = numNodes;

        testConf.nodeData[0][0] = numNodes;
        testConf.nodeData[0][1] = Double.parseDouble(splits[1]);

        move();
        match(NODE_DATA_END);
    }

    private void parseBranchData(CaseFileData testConf){
        move();
        ArrayList<String> branchDataList = collectLines(BRANCH_DATA_END);

        int iBranchNumber = branchDataList.size();
        testConf.branchData = new Object[iBranchNumber][5];
        testConf.iBranchData = iBranchNumber;

        for (int i = 0; i < iBranchNumber; i++) {
            String strBranch = (String) branchDataList.get(i);
            int iBranchFields = 4;

            while (iBranchFields > 0) {
                int iIndex = strBranch.lastIndexOf("\t");
                if (iIndex < 0) {
                    iIndex = strBranch.lastIndexOf(" ");
                }

                String strData = strBranch.substring(iIndex + 1);
                strData = strData.trim();

                if (iBranchFields > 2) {
                    double dTemp = Double.parseDouble(strData);
                    testConf.branchData[i][iBranchFields] = String.format("%1$15.4f", dTemp);
                } else {
                    testConf.branchData[i][iBranchFields] = Integer.parseInt(strData);
                }

                iBranchFields--;

                if (iIndex > 0) {
                    strBranch = strBranch.substring(0, iIndex);
                }
            }

            if (strBranch.length() > 0) {
                testConf.branchData[i][0] = strBranch;
            }
        }
    }

    /**
     *
     * Collect all of the lines in the GenData section and create GenData
     * object from the fields.
     *
     * @param testConf
     * @throws BadDataFileFormatException
     */
    private void parseGenData(CaseFileData testConf) throws BadDataFileFormatException{
        move();
        ArrayList<String> genDataList = collectLines(GEN_DATA_END);

        int iGenNumber = genDataList.size();
        testConf.genData = new GenData[iGenNumber];
        testConf.iGenData = iGenNumber;

        for (int i = 0; i < iGenNumber; i++) {
            String strGen = (String) genDataList.get(i);

            //Split the line at whit space.
            String[] lineElems = strGen.split(WS_REG_EX);
            if (lineElems.length != 9){
              throw new BadDataFileFormatException(strGen);
            }

            try{
            testConf.genData[i] = new GenData(
                lineElems[0],//name
                Integer.parseInt(lineElems[1]),//id
                Integer.parseInt(lineElems[2]),//atBus
                Double.parseDouble(lineElems[3]),//sCost
                Double.parseDouble(lineElems[4]),//a
                Double.parseDouble(lineElems[5]),//b
                Double.parseDouble(lineElems[6]),//capL
                Double.parseDouble(lineElems[7]),//capU
                Double.parseDouble(lineElems[8])//initMoney
                );
            }catch(NumberFormatException nfe){
              throw new BadDataFileFormatException(nfe);
            }
        }
    }

    private void parseLSEFixedDemand(CaseFileData testConf){
        move();
        ArrayList<String> LSEDataList = collectLines(LSE_DATA_FIXED_DEM_END);

        int iLSENumber = LSEDataList.size() / 3;
        testConf.lseSec1Data = new Object[iLSENumber][11];
        testConf.lseSec2Data = new Object[iLSENumber][11];
        testConf.lseSec3Data = new Object[iLSENumber][11];
        testConf.lseData = new Object[iLSENumber][27];
        testConf.iLSEData = iLSENumber;

        for (int i = 0; i < iLSENumber * 3; i++) {
            String strLSE = (String) LSEDataList.get(i);
            int iLSEFields = 10;
            int iLSEIndex = i % iLSENumber;

            while (iLSEFields > 0) {
                int iIndex = strLSE.lastIndexOf("\t");
                if (iIndex < 0) {
                    iIndex = strLSE.lastIndexOf(" ");
                }

                String strData = strLSE.substring(iIndex + 1);
                strData = strData.trim();

                if (i < iLSENumber) {
                    if (iLSEFields > 2) {
                        double dTemp = Double.parseDouble(strData);
                        testConf.lseSec1Data[iLSEIndex][iLSEFields] = String.format("%1$15.4f", dTemp);
                        testConf.lseData[iLSEIndex][iLSEFields] = String.format("%1$15.4f", dTemp);
                    } else {
                        testConf.lseSec1Data[iLSEIndex][iLSEFields] = Integer.parseInt(strData);
                        testConf.lseData[iLSEIndex][iLSEFields] = Integer.parseInt(strData);
                    }
                } else if ((i >= iLSENumber) && (i < 2 * iLSENumber)) {
                    if (iLSEFields > 2) {
                        double dTemp = Double.parseDouble(strData);
                        testConf.lseSec2Data[iLSEIndex][iLSEFields] = String.format("%1$15.4f", dTemp);
                        testConf.lseData[iLSEIndex][iLSEFields + 8] = String.format("%1$15.4f", dTemp);
                    } else {
                        testConf.lseSec2Data[iLSEIndex][iLSEFields] = Integer.parseInt(strData);
                    }

                } else if (i >= 2 * iLSENumber) {
                    if (iLSEFields > 2) {
                        double dTemp = Double.parseDouble(strData);
                        testConf.lseSec3Data[iLSEIndex][iLSEFields] = String.format("%1$15.4f", dTemp);
                        testConf.lseData[iLSEIndex][iLSEFields + 16] = String.format("%1$15.4f", dTemp);
                    } else {
                        testConf.lseSec3Data[iLSEIndex][iLSEFields] = Integer.parseInt(strData);
                    }
                }

                iLSEFields--;

                if (iIndex > 0) {
                    strLSE = strLSE.substring(0, iIndex);
                }
            }

            if (i < iLSENumber) {
                if (strLSE.length() > 0) {
                    testConf.lseSec1Data[iLSEIndex][0] = strLSE;
                    testConf.lseData[iLSEIndex][0] = strLSE;
                }
            } else if ((i >= iLSENumber) && (i < 2 * iLSENumber)) {
                if (strLSE.length() > 0) {
                    testConf.lseSec2Data[iLSEIndex][0] = strLSE;
                }
            } else if (i >= 2 * iLSENumber) {
                if (strLSE.length() > 0) {
                    testConf.lseSec3Data[iLSEIndex][0] = strLSE;
                }
            }
        }
    }

    private void parseLSEPSensDemand(CaseFileData testConf){
        move();
        ArrayList<String> LSEPriceDemandDataList = collectLines(LSE_DATA_PRICE_SENS_DEM_END);

        int iLSEDemandNumber = LSEPriceDemandDataList.size() / 24;
        testConf.lsePriceSensitiveDemand = new Object[iLSEDemandNumber][24][7];
        testConf.iLSEData = iLSEDemandNumber;

        for (int i = 0; i < iLSEDemandNumber; i++) {
            for (int j = 0; j < 24; j++) {
                String strLSEDemand = (String) LSEPriceDemandDataList.get(i * 24 + j);
                int iDemandFields = 6;

                while (iDemandFields > 0) {
                    int iIndex = strLSEDemand.lastIndexOf("\t");
                    if (iIndex < 0) {
                        iIndex = strLSEDemand.lastIndexOf(" ");
                    }

                    String strData = strLSEDemand.substring(iIndex + 1);
                    strData = strData.trim();

                    if (iDemandFields > 3) {
                        testConf.lsePriceSensitiveDemand[i][j][iDemandFields] = Double.parseDouble(strData);
                    } else {
                        testConf.lsePriceSensitiveDemand[i][j][iDemandFields] = Integer.parseInt(strData);
                    }

                    iDemandFields--;

                    if (iIndex > 0) {
                        strLSEDemand = strLSEDemand.substring(0, iIndex);
                    }
                }

                if (strLSEDemand.length() > 0) {
                    testConf.lsePriceSensitiveDemand[i][j][0] = strLSEDemand;
                }
            }
        }
    }

    private void parseLSEHybDemand(CaseFileData testConf){
        move();
        ArrayList<String> LSEHybridDemandDataList = collectLines(LSE_DATA_HYBRID_DEM_END);

        int iLSENumber = LSEHybridDemandDataList.size() / 3;
        testConf.lseHybridDemand = new Object[iLSENumber][27];
        testConf.iLSEData = iLSENumber;

        for (int i = 0; i < iLSENumber * 3; i++) {
            String strLSE = (String) LSEHybridDemandDataList.get(i);
            int iLSEFields = 10;
            int iLSEIndex = i % iLSENumber;

            while (iLSEFields > 0) {
                int iIndex = strLSE.lastIndexOf("\t");
                if (iIndex < 0) {
                    iIndex = strLSE.lastIndexOf(" ");
                }

                String strData = strLSE.substring(iIndex + 1);
                strData = strData.trim();

                if (i < iLSENumber) {
                    if (iLSEFields > 0) {
                        testConf.lseHybridDemand[iLSEIndex][iLSEFields] = Integer.parseInt(strData);
                    }
                } else if ((i >= iLSENumber) && (i < 2 * iLSENumber)) {
                    if (iLSEFields > 2) {
                        testConf.lseHybridDemand[iLSEIndex][iLSEFields + 8] = Integer.parseInt(strData);
                    }
                } else if (i >= 2 * iLSENumber) {
                    if (iLSEFields > 2) {
                        testConf.lseHybridDemand[iLSEIndex][iLSEFields + 16] = Integer.parseInt(strData);
                    }
                }

                iLSEFields--;

                if (iIndex > 0) {
                    strLSE = strLSE.substring(0, iIndex);
                }
            }

            if (i < iLSENumber) {
                if (strLSE.length() > 0) {
                    testConf.lseHybridDemand[iLSEIndex][0] = strLSE;
                }
            } else if ((i >= iLSENumber) && (i < 2 * iLSENumber)) {
                if (strLSE.length() > 0) {
                    testConf.lseHybridDemand[iLSEIndex][0] = strLSE;
                }
            } else if (i >= 2 * iLSENumber) {
                if (strLSE.length() > 0) {
                    testConf.lseHybridDemand[iLSEIndex][0] = strLSE;
                }
            }
        }
    }

    private void parseGenLearningData(CaseFileData testConf){
        move();
        ArrayList<String> genLearningDataList = collectLines(GEN_LEARNING_DATA_END);

        int iGenNumber = genLearningDataList.size();
        testConf.genLearningData = new double[iGenNumber][12];
        testConf.iGenData = iGenNumber;

        for (int i = 0; i < iGenNumber; i++) {
            String strGenLearning = (String) genLearningDataList.get(i);
            int iGenLearningFields = 11;

            while (iGenLearningFields >= 0) {
                int iIndex = strGenLearning.lastIndexOf("\t");
                if (iIndex < 0) {
                    iIndex = strGenLearning.lastIndexOf(" ");
                }

                String strData = strGenLearning.substring(iIndex + 1);
                strData = strData.trim();

                double dTemp = Double.parseDouble(strData);
                testConf.genLearningData[i][iGenLearningFields] = dTemp;

                iGenLearningFields--;

                if (iIndex >= 0) {
                    strGenLearning = strGenLearning.substring(0, iIndex);
                }
            }

        }
    }

    private void loadDefaultGenLearningData(CaseFileData testConf){
        DefaultSimulationParameters defSimParams = new DefaultSimulationParameters();
        testConf.genLearningData = new double[testConf.iGenData][12];

        for (int i = 0; i < testConf.iGenData; i++) {
            testConf.genLearningData[i][0] = defSimParams.Default_InitPropensity;
            testConf.genLearningData[i][1] = defSimParams.Default_Cooling;
            testConf.genLearningData[i][2] = defSimParams.Default_Recency;
            testConf.genLearningData[i][3] = defSimParams.Default_Experimentation;
            testConf.genLearningData[i][4] = defSimParams.Default_M1;
            testConf.genLearningData[i][5] = defSimParams.Default_M2;
            testConf.genLearningData[i][6] = defSimParams.Default_M3;
            testConf.genLearningData[i][7] = defSimParams.Default_RI_MAX_Lower;
            testConf.genLearningData[i][8] = defSimParams.Default_RI_MAX_Upper;
            testConf.genLearningData[i][9] = defSimParams.Default_RI_MIN_C;
            testConf.genLearningData[i][10] = defSimParams.Default_SlopeStart;
            testConf.genLearningData[i][11] = defSimParams.Default_iRewardSelection;
        }
    }

    private ArrayList<String> collectLines(String endMarker){
        ArrayList<String> lines = new ArrayList<String>();

        while(!endMarker.equals(currentLine)){
            lines.add(currentLine);
            move();
        }

        return lines;
    }

    /**
     * Load the case file data.
     *
     * Assumes the internal inputReader has been initialized.
     * @throws BadDataFileFormatException
     */
    private CaseFileData loadCaseFileData() throws BadDataFileFormatException{
      CaseFileData testConf = new CaseFileData();

      parseDataFile(testConf);

      inputReader.close();
      return testConf;
    }

    /**
     * A simple line reader which skips any line which
     * starts with a comment delimiter.
     */
    private static class SimpleLineReader {

        Scanner scanner;

        public SimpleLineReader(File inputFile) throws FileNotFoundException{
            scanner = new Scanner(inputFile);
        }

        public SimpleLineReader(Reader inputReader){
          scanner = new Scanner(inputReader);
        }


        String nextLine(){
            String line = "";
            boolean scanMore = false;
            do{
                scanMore = false;
                if(!scanner.hasNext()) {
                    line = null;
                    break;
                }

                line = scanner.nextLine();

                line = line.trim();
                if (line.length() == 0) {
                    scanMore = true;
                }

                int iCommentIndex = line.indexOf("//");
                if (iCommentIndex == 0) {
                    scanMore = true;
                }
            }while(scanMore);

            return line;
        }

        boolean hasNext(){
            return scanner.hasNext();
        }

        void close(){
            scanner.close();
        }
    }

    /**
     * An exception to represent a problem in the data file.
     *
     * @author Sean L. Mooney
     *
     */
    public static class BadDataFileFormatException extends Exception{
        public BadDataFileFormatException(String invalidLine){
            super(invalidLine);
        }

        public BadDataFileFormatException(Throwable cause){
          super(cause);
        }
    }
}
